"""
Zappa Asynchronous Tasks

Example:
```
from zappa.asynchronous import task

@task(service='sns')
def my_async_func(*args, **kwargs):
    dosomething()
```

For SNS, you can also pass an `arn` argument to task() which will specify which SNS path to send it to.

Without `service='sns'`, the default service is 'lambda' which will call the method in an asynchronous
lambda call.

The following restrictions apply:
* function must have a clean import path -- i.e. no closures, lambdas, or methods.
* args and kwargs must be JSON-serializable.
* The JSON-serialized form must be within the size limits for Lambda (128K) or SNS (256K) events.

Discussion of this comes from:
    https://github.com/Miserlou/Zappa/issues/61
    https://github.com/Miserlou/Zappa/issues/603
    https://github.com/Miserlou/Zappa/pull/694
    https://github.com/Miserlou/Zappa/pull/732
    https://github.com/Miserlou/Zappa/issues/840

## Full lifetime of an asynchronous dispatch:

1. In a file called `foo.py`, there is the following code:

```
   from zappa.asynchronous import task

   @task
   def my_async_func(*args, **kwargs):
       return sum(args)
```

2. The decorator desugars to:
   `my_async_func = task(my_async_func)`

3. Somewhere else, the code runs:
   `res = my_async_func(1,2)`
   really calls task's `_run_async(1,2)`
      with `func` equal to the original `my_async_func`
   If we are running in Lambda, this runs:
      LambdaAsyncResponse().send('foo.my_async_func', (1,2), {})
   and returns the LambdaAsyncResponse instance to the local
   context.  That local context, can, e.g. test for `res.sent`
   to confirm it was dispatched correctly.

4. LambdaAsyncResponse.send invoked the currently running
   AWS Lambda instance with the json message:

```
   { "command": "zappa.asynchronous.route_lambda_task",
     "task_path": "foo.my_async_func",
     "args": [1,2],
     "kwargs": {}
   }
```

5. The new lambda instance is invoked with the message above,
   and Zappa runs its usual bootstrapping context, and inside
   zappa.handler, the existence of the 'command' key in the message
   dispatches the full message to zappa.asynchronous.route_lambda_task, which
   in turn calls `run_message(message)`

6. `run_message` loads the task_path value to load the `func` from `foo.py`.
   We should note that my_async_func is wrapped by @task in this new
   context, as well.  However, @task also decorated `my_async_func.sync()`
   to run the original function synchronously.

   `run_message` duck-types the method and finds the `.sync` attribute
   and runs that instead -- thus we do not infinitely dispatch.

   If `my_async_func` had code to dispatch other functions inside its
   synchronous portions (or even call itself recursively), those *would*
   be dispatched asynchronously, unless, of course, they were called
   by: `my_async_func.sync(1,2)` in which case it would run synchronously
   and in the current lambda function.

"""

import importlib
import inspect
import json
import os
import time
import uuid
from functools import update_wrapper, wraps

import boto3
import botocore

from .utilities import get_topic_name, validate_json_serializable

try:
    from zappa_settings import ASYNC_RESPONSE_TABLE
except ImportError:
    ASYNC_RESPONSE_TABLE = None

# Declare these here so they're kept warm.
try:
    aws_session = boto3.Session()
    LAMBDA_CLIENT = aws_session.client("lambda")
    SNS_CLIENT = aws_session.client("sns")
    STS_CLIENT = aws_session.client("sts")
    DYNAMODB_CLIENT = aws_session.client("dynamodb")
except botocore.exceptions.NoRegionError:  # pragma: no cover
    # This can happen while testing on Travis, but it's taken care  of
    # during class initialization.
    pass


##
# Response and Exception classes
##

LAMBDA_ASYNC_PAYLOAD_LIMIT = 256000
SNS_ASYNC_PAYLOAD_LIMIT = 256000


class AsyncException(Exception):  # pragma: no cover
    """Simple exception class for async tasks."""

    pass


class LambdaAsyncResponse:
    """
    Base Response Dispatcher class
    Can be used directly or subclassed if the method to send the message is changed.
    """

    def __init__(self, lambda_function_name=None, aws_region=None, capture_response=False, **kwargs):
        """ """
        if kwargs.get("boto_session"):
            self.client = kwargs.get("boto_session").client("lambda")
        else:  # pragma: no cover
            self.client = LAMBDA_CLIENT

        self.lambda_function_name = lambda_function_name
        self.aws_region = aws_region
        if capture_response:
            if ASYNC_RESPONSE_TABLE is None:
                print(
                    "Warning! Attempted to capture a response without "
                    "async_response_table configured in settings (you won't "
                    "capture async responses)."
                )
                capture_response = False
                self.response_id = "MISCONFIGURED"

            else:
                self.response_id = str(uuid.uuid4())
        else:
            self.response_id = None

        self.capture_response = capture_response

    def send(self, task_path, args, kwargs):
        """
        Create the message object and pass it to the actual sender.
        """
        message = {
            "task_path": task_path,
            "capture_response": self.capture_response,
            "response_id": self.response_id,
            "args": args,
            "kwargs": kwargs,
        }
        self._send(message)
        return self

    def _send(self, message):
        """
        Given a message, directly invoke the lamdba function for this task.
        """
        message["command"] = "zappa.asynchronous.route_lambda_task"
        payload = json.dumps(message)
        if len(payload) > LAMBDA_ASYNC_PAYLOAD_LIMIT:  # pragma: no cover
            raise AsyncException("Payload too large for async Lambda call")
        self.response = self.client.invoke(
            FunctionName=self.lambda_function_name,
            InvocationType="Event",  # makes the call async
            Payload=payload,
        )
        self.sent = self.response.get("StatusCode", 0) == 202


class SnsAsyncResponse(LambdaAsyncResponse):
    """
    Send a SNS message to a specified SNS topic
    Serialise the func path and arguments
    """

    def __init__(self, lambda_function_name=None, aws_region=None, capture_response=False, **kwargs):
        self.lambda_function_name = lambda_function_name
        self.aws_region = aws_region

        if kwargs.get("boto_session"):
            self.client = kwargs.get("boto_session").client("sns")
        else:  # pragma: no cover
            self.client = SNS_CLIENT

        if kwargs.get("arn"):
            self.arn = kwargs.get("arn")
        else:
            if kwargs.get("boto_session"):
                sts_client = kwargs.get("boto_session").client("sts")
            else:
                sts_client = STS_CLIENT
            AWS_ACCOUNT_ID = sts_client.get_caller_identity()["Account"]
            self.arn = "arn:aws:sns:{region}:{account}:{topic_name}".format(
                region=self.aws_region,
                account=AWS_ACCOUNT_ID,
                topic_name=get_topic_name(self.lambda_function_name),
            )

        # Issue: https://github.com/Miserlou/Zappa/issues/1209
        # TODO: Refactor
        self.capture_response = capture_response
        if capture_response:
            if ASYNC_RESPONSE_TABLE is None:
                print(
                    "Warning! Attempted to capture a response without "
                    "async_response_table configured in settings (you won't "
                    "capture async responses)."
                )
                capture_response = False
                self.response_id = "MISCONFIGURED"

            else:
                self.response_id = str(uuid.uuid4())
        else:
            self.response_id = None

        self.capture_response = capture_response

    def _send(self, message):
        """
        Given a message, publish to this topic.
        """
        message["command"] = "zappa.asynchronous.route_sns_task"
        payload = json.dumps(message)
        if len(payload) > LAMBDA_ASYNC_PAYLOAD_LIMIT:  # pragma: no cover
            raise AsyncException("Payload too large for SNS")
        self.response = self.client.publish(TargetArn=self.arn, Message=payload)
        self.sent = self.response.get("MessageId")


##
# Aync Routers
##

ASYNC_CLASSES = {
    "lambda": LambdaAsyncResponse,
    "sns": SnsAsyncResponse,
}


def route_lambda_task(event, context):
    """
    Deserialises the message from event passed to zappa.handler.run_function
    imports the function, calls the function with args
    """
    message = event
    return run_message(message)


def route_sns_task(event, context):
    """
    Gets SNS Message, deserialises the message,
    imports the function, calls the function with args
    """
    record = event["Records"][0]
    message = json.loads(record["Sns"]["Message"])
    return run_message(message)


def run_message(message):
    """
    Runs a function defined by a message object with keys:
    'task_path', 'args', and 'kwargs' used by lambda routing
    and a 'command' in handler.py
    """
    if message.get("capture_response", False):
        DYNAMODB_CLIENT.put_item(
            TableName=ASYNC_RESPONSE_TABLE,
            Item={
                "id": {"S": str(message["response_id"])},
                "ttl": {"N": str(int(time.time() + 600))},
                "async_status": {"S": "in progress"},
                "async_response": {"S": str(json.dumps("N/A"))},
            },
        )

    func = import_and_get_task(message["task_path"])
    if hasattr(func, "sync"):
        response = func.sync(*message["args"], **message["kwargs"])
    else:
        response = func(*message["args"], **message["kwargs"])

    if message.get("capture_response", False):
        DYNAMODB_CLIENT.update_item(
            TableName=ASYNC_RESPONSE_TABLE,
            Key={"id": {"S": str(message["response_id"])}},
            UpdateExpression="SET async_response = :r, async_status = :s",
            ExpressionAttributeValues={
                ":r": {"S": str(json.dumps(response))},
                ":s": {"S": "complete"},
            },
        )

    return response


##
# Execution interfaces and classes
##


def run(
    func,
    args=[],
    kwargs={},
    service="lambda",
    capture_response=False,
    remote_aws_lambda_function_name=None,
    remote_aws_region=None,
    **task_kwargs
):
    """
    Instead of decorating a function with @task, you can just run it directly.
    If you were going to do func(*args, **kwargs), then you will call this:

    import zappa.asynchronous.run
    zappa.asynchronous.run(func, args, kwargs)

    If you want to use SNS, then do:

    zappa.asynchronous.run(func, args, kwargs, service='sns')

    and other arguments are similar to @task
    """
    lambda_function_name = remote_aws_lambda_function_name or os.environ.get("AWS_LAMBDA_FUNCTION_NAME")
    aws_region = remote_aws_region or os.environ.get("AWS_REGION")

    task_path = get_func_task_path(func)
    return ASYNC_CLASSES[service](
        lambda_function_name=lambda_function_name, aws_region=aws_region, capture_response=capture_response, **task_kwargs
    ).send(task_path, args, kwargs)


# Handy:
# http://stackoverflow.com/questions/10294014/python-decorator-best-practice-using-a-class-vs-a-function
# However, this needs to pass inspect.getargspec() in handler.py which does not take classes
# Wrapper written to take optional arguments
# http://chase-seibert.github.io/blog/2013/12/17/python-decorator-optional-parameter.html
def task(*args, **kwargs):
    """Async task decorator so that running

    Args:
        func (function): the function to be wrapped
            Further requirements:
            func must be an independent top-level function.
                 i.e. not a class method or an anonymous function
        service (str): either 'lambda' or 'sns'
        remote_aws_lambda_function_name (str): the name of a remote lambda function to call with this task
        remote_aws_region (str): the name of a remote region to make lambda/sns calls against

    Returns:
        A replacement function that dispatches func() to
        run asynchronously through the service in question
    """

    func = None
    if len(args) == 1 and callable(args[0]):
        func = args[0]

    if not kwargs:  # Default Values
        service = "lambda"
        lambda_function_name_arg = None
        aws_region_arg = None

    else:  # Arguments were passed
        service = kwargs.get("service", "lambda")
        lambda_function_name_arg = kwargs.get("remote_aws_lambda_function_name")
        aws_region_arg = kwargs.get("remote_aws_region")

    capture_response = kwargs.get("capture_response", False)

    def func_wrapper(func):
        task_path = get_func_task_path(func)

        @wraps(func)
        def _run_async(*args, **kwargs):
            """
            This is the wrapping async function that replaces the function
            that is decorated with @task.
            Args:
                These are just passed through to @task's func

            Assuming a valid service is passed to task() and it is run
            inside a Lambda process (i.e. AWS_LAMBDA_FUNCTION_NAME exists),
            it dispatches the function to be run through the service variable.
            Otherwise, it runs the task synchronously.

            Returns:
                In async mode, the object returned includes state of the dispatch.
                For instance

                When outside of Lambda, the func passed to @task is run and we
                return the actual value.
            """
            lambda_function_name = lambda_function_name_arg or os.environ.get("AWS_LAMBDA_FUNCTION_NAME")
            aws_region = aws_region_arg or os.environ.get("AWS_REGION")

            if (service in ASYNC_CLASSES) and (lambda_function_name):
                send_result = ASYNC_CLASSES[service](
                    lambda_function_name=lambda_function_name,
                    aws_region=aws_region,
                    capture_response=capture_response,
                ).send(task_path, args, kwargs)
                return send_result
            else:
                validate_json_serializable(*args, **kwargs)
                return func(*args, **kwargs)

        update_wrapper(_run_async, func)

        _run_async.service = service
        _run_async.sync = func

        return _run_async

    return func_wrapper(func) if func else func_wrapper


def task_sns(func):
    """
    SNS-based task dispatcher. Functions the same way as task()
    """
    return task(func, service="sns")


##
# Utility Functions
##


def import_and_get_task(task_path):
    """
    Given a modular path to a function, import that module
    and return the function.
    """
    module, function = task_path.rsplit(".", 1)
    app_module = importlib.import_module(module)
    app_function = getattr(app_module, function)
    return app_function


def get_func_task_path(func):
    """
    Format the modular task path for a function via inspection.
    """
    module_path = inspect.getmodule(func).__name__
    task_path = "{module_path}.{func_name}".format(module_path=module_path, func_name=func.__name__)
    return task_path


def get_async_response(response_id):
    """
    Get the response from the async table
    """
    response = DYNAMODB_CLIENT.get_item(TableName=ASYNC_RESPONSE_TABLE, Key={"id": {"S": str(response_id)}})
    if "Item" not in response:
        return None

    return {
        "status": response["Item"]["async_status"]["S"],
        "response": json.loads(response["Item"]["async_response"]["S"]),
    }



import base64
import logging
import sys
from io import BytesIO
from typing import Optional
from urllib.parse import unquote, urlencode

from .utilities import ApacheNCSAFormatter, merge_headers, titlecase_keys

BINARY_METHODS = ["POST", "PUT", "PATCH", "DELETE", "CONNECT", "OPTIONS"]


logger = logging.getLogger(__name__)


def create_wsgi_request(
    event_info,
    server_name="zappa",
    script_name=None,
    trailing_slash=True,
    binary_support=False,
    base_path=None,
    context_header_mappings={},
):
    """
    Given some event_info via API Gateway,
    create and return a valid WSGI request environ.
    """
    if event_info.get("version", "") == "2.0":
        method, headers, path, query_string, remote_user, authorizer = process_lambda_payload_v2(event_info)
    else:
        method, headers, path, query_string, remote_user, authorizer = process_lambda_payload_v1(event_info)

    if context_header_mappings:
        for key, value in context_header_mappings.items():
            parts = value.split(".")
            header_val = event_info["requestContext"]
            for part in parts:
                if part not in header_val:
                    header_val = None
                    break
                else:
                    header_val = header_val[part]
            if header_val is not None:
                headers[key] = header_val

    # Related:  https://github.com/Miserlou/Zappa/issues/677
    #           https://github.com/Miserlou/Zappa/issues/683
    #           https://github.com/Miserlou/Zappa/issues/696
    #           https://github.com/Miserlou/Zappa/issues/836
    #           https://en.wikipedia.org/wiki/Hypertext_Transfer_Protocol#Summary_table
    if binary_support and (method in BINARY_METHODS):
        if event_info.get("isBase64Encoded", False):
            encoded_body = event_info["body"]
            body = base64.b64decode(encoded_body)
        else:
            body = event_info.get("body")
            if isinstance(body, str):
                body = body.encode("utf-8")

    else:
        body = event_info.get("body")
        if isinstance(body, str):
            body = body.encode("utf-8")

    # Make header names canonical, e.g. content-type => Content-Type
    # https://github.com/Miserlou/Zappa/issues/1188
    headers = titlecase_keys(headers)

    if base_path:
        script_name = f"/{base_path}"

        if path.startswith(script_name):
            path = path[len(script_name) :]

    x_forwarded_for = headers.get("X-Forwarded-For", "")
    if "," in x_forwarded_for:
        # The last one is the cloudfront proxy ip. The second to last is the real client ip.
        # Everything else is user supplied and untrustworthy.
        addresses = [addr.strip() for addr in x_forwarded_for.split(",")]
        remote_addr = addresses[-2]
    else:
        remote_addr = x_forwarded_for or "127.0.0.1"

    environ = {
        "PATH_INFO": get_wsgi_string(path),
        "QUERY_STRING": get_wsgi_string(query_string),
        "REMOTE_ADDR": remote_addr,
        "REQUEST_METHOD": method,
        "SCRIPT_NAME": get_wsgi_string(str(script_name)) if script_name else "",
        "SERVER_NAME": str(server_name),
        "SERVER_PORT": headers.get("X-Forwarded-Port", "80"),
        "SERVER_PROTOCOL": str("HTTP/1.1"),
        "wsgi.version": (1, 0),
        "wsgi.url_scheme": headers.get("X-Forwarded-Proto", "http"),
        # This must be Bytes or None
        # - https://docs.djangoproject.com/en/4.2/releases/4.2/#miscellaneous
        # - https://wsgi.readthedocs.io/en/latest/definitions.html#envvar-wsgi.input
        # > Manually instantiated WSGIRequest objects must be provided
        # > a file-like object for wsgi.input.
        "wsgi.input": BytesIO(body),
        "wsgi.errors": sys.stderr,
        "wsgi.multiprocess": False,
        "wsgi.multithread": False,
        "wsgi.run_once": False,
    }

    if authorizer:
        environ["API_GATEWAY_AUTHORIZER"] = authorizer

    # Input processing
    if method in ["POST", "PUT", "PATCH", "DELETE"]:
        if "Content-Type" in headers:
            environ["CONTENT_TYPE"] = headers["Content-Type"]

        if body:
            environ["CONTENT_LENGTH"] = str(len(body))
        else:
            environ["CONTENT_LENGTH"] = "0"

    for header in headers:
        wsgi_name = "HTTP_" + header.upper().replace("-", "_")
        environ[wsgi_name] = str(headers[header])

    if script_name:
        environ["SCRIPT_NAME"] = script_name
        # Strip the stage prefix from PATH_INFO if it exists.
        # This is required for API Gateway v2 where rawPath includes the stage (e.g., /stage/path).
        # For v1, this is a no-op since AWS already strips the stage from event["path"].
        # Note: This fixes a bug in master where .replace() was called without assignment.
        path_info = environ["PATH_INFO"]
        if path_info.startswith(script_name):
            environ["PATH_INFO"] = path_info[len(script_name) :]

    if remote_user:
        environ["REMOTE_USER"] = remote_user

    return environ


def process_lambda_payload_v1(event_info):
    method = event_info.get("httpMethod", None)
    headers = merge_headers(event_info) or {}  # Allow for the AGW console 'Test' button to work (Pull #735)
    path = unquote(event_info["path"])
    # API Gateway and ALB both started allowing for multi-value querystring
    # params in Nov. 2018. If there aren't multi-value params present, then
    # it acts identically to 'queryStringParameters', so we can use it as a
    # drop-in replacement.
    #
    # The one caveat here is that ALB will only include _one_ of
    # queryStringParameters _or_ multiValueQueryStringParameters, which means
    # we have to check for the existence of one and then fall back to the
    # other.
    if "multiValueQueryStringParameters" in event_info:
        query = event_info["multiValueQueryStringParameters"]
        query_string = urlencode(query, doseq=True) if query else ""
    else:
        query = event_info.get("queryStringParameters", {})
        query_string = urlencode(query) if query else ""
    # Systems calling the Lambda (other than API Gateway) may not provide the field requestContext
    # Extract remote_user, authorizer if Authorizer is enabled
    authorizer, remote_user = None, None
    if "requestContext" in event_info:
        authorizer = event_info["requestContext"].get("authorizer", None)
        if authorizer:
            remote_user = authorizer.get("principalId")
        elif event_info["requestContext"].get("identity"):
            remote_user = event_info["requestContext"]["identity"].get("userArn")
    return method, headers, path, query_string, remote_user, authorizer


def process_lambda_payload_v2(event_info):
    # See the new format documentation
    # here: https://docs.aws.amazon.com/lambda/latest/dg/urls-invocation.html#urls-payloads
    method = event_info["requestContext"]["http"]["method"]
    headers = event_info["headers"]
    if event_info.get("cookies"):
        headers["cookie"] = "; ".join(event_info["cookies"])
    path = unquote(event_info["rawPath"])
    query = event_info.get("queryStringParameters", {})
    query_string = urlencode(query) if query else ""
    # Systems calling the Lambda (other than API Gateway) may not provide the field requestContext
    # Extract remote_user, authorizer if Authorizer is enabled
    remote_user = None
    authorizer = event_info["requestContext"].get("authorizer", None)
    if authorizer:
        if authorizer.get("lambda"):
            # Need to add principalId manually to lambda authorizer response context
            remote_user = authorizer["lambda"].get("principalId")
        elif authorizer.get("iam"):
            remote_user = authorizer["iam"].get("userArn")
    return method, headers, path, query_string, remote_user, authorizer


def common_log(environ, response, response_time: Optional[int] = None):
    """
    Given the WSGI environ and the response,
    log this event in Common Log Format.

    response_time: response time in micro-seconds
    """

    logger = logging.getLogger(__name__)

    if response_time:
        formatter = ApacheNCSAFormatter(with_response_time=True)
        log_entry = formatter(
            response.status_code,
            environ,
            len(response.content),
            rt_us=response_time,
        )
    else:
        formatter = ApacheNCSAFormatter(with_response_time=False)
        log_entry = formatter(response.status_code, environ, len(response.content))

    logger.info(log_entry)

    return log_entry


# Related: https://github.com/Miserlou/Zappa/issues/1199
def get_wsgi_string(string, encoding="utf-8"):
    """
    Returns wsgi-compatible string
    """
    return string.encode(encoding).decode("iso-8859-1")



import unittest
from typing import Any, Tuple

from mock import Mock

from zappa.handler import LambdaHandler
from zappa.utilities import merge_headers

from .utils import is_base64


def no_args() -> None:
    return


def one_arg(first: Any) -> Any:
    return first


def two_args(first: Any, second: Any) -> Tuple[Any, Any]:
    return first, second


def var_args(*args: Any) -> Tuple[Any, ...]:
    return args


def var_args_with_one(first: Any, *args: Any) -> Tuple[Any, Any]:
    return first, args[0]


def unsupported(first: Any, second: Any, third: Any) -> Tuple[Any, Any, Any]:
    return first, second, third


def raises_exception(*args: Any, **kwargs: Any) -> None:
    raise Exception("app exception")


def handle_bot_intent(event: Any, context: Any) -> str:
    return "Success"


mocked_exception_handler = Mock()


class TestZappa(unittest.TestCase):
    def setUp(self):
        mocked_exception_handler.reset_mock()

    def tearDown(self):
        LambdaHandler._LambdaHandler__instance = None
        LambdaHandler.settings = None
        LambdaHandler.settings_name = None

    def test_run_function(self):
        self.assertIsNone(LambdaHandler.run_function(no_args, "e", "c"))
        self.assertEqual(LambdaHandler.run_function(one_arg, "e", "c"), "e")
        self.assertEqual(LambdaHandler.run_function(two_args, "e", "c"), ("e", "c"))
        self.assertEqual(LambdaHandler.run_function(var_args, "e", "c"), ("e", "c"))
        self.assertEqual(LambdaHandler.run_function(var_args_with_one, "e", "c"), ("e", "c"))

        try:
            LambdaHandler.run_function(unsupported, "e", "c")
            self.fail("Exception expected")
        except RuntimeError as e:
            pass

    def test_run_fuction_with_type_hint(self):
        scope = {}
        exec("def f_with_type_hint() -> None: return", scope)
        f_with_type_hint = scope["f_with_type_hint"]
        self.assertIsNone(LambdaHandler.run_function(f_with_type_hint, "e", "c"))

    def test_wsgi_script_name_on_aws_url(self):
        """
        Ensure that requests to the amazonaws.com host for an API with a
        domain have the correct request.url
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/return/request/url",
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertEqual(
            response["body"],
            "https://1234567890.execute-api.us-east-1.amazonaws.com/dev/return/request/url",
        )

    def test_wsgi_script_name_on_domain_url(self):
        """
        Ensure that requests to the amazonaws.com host for an API with a
        domain have the correct request.url
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "example.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/return/request/url",
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertEqual(response["body"], "https://example.com/return/request/url")

    def test_wsgi_script_name_with_multi_value_header(self):
        """
        Ensure that requests generated with multivalued headers (such as
        from an ALB with Multi Valued Headers enabled) succeed.
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "multiValueHeaders": {
                "Host": ["example.com"],
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/return/request/url",
        }
        response = lh.handler(event, None)
        self.assertEqual(response["statusCode"], 200)
        self.assertIn("multiValueHeaders", response)

    def test_wsgi_script_name_with_multi_value_querystring(self):
        """
        Ensure that requests generated with multivalue querystrings succeed.
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "multiValueQueryStringParameters": {"multi": ["value", "qs"]},
            "multiValueHeaders": {
                "Host": ["example.com"],
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/return/request/url",
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertEqual(
            response["body"],
            "https://example.com/return/request/url?multi=value&multi=qs",
        )

    def test_wsgi_script_name_on_test_request(self):
        """
        Ensure that requests sent by the "Send test request" button behaves
        sensibly
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {},
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/return/request/url",
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertEqual(response["body"], "https://zappa:80/return/request/url")

    def test_exception_handler_on_web_request(self):
        """
        Ensure that app exceptions triggered by web requests use the exception_handler.
        """
        lh = LambdaHandler("tests.test_exception_handler_settings")

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/return/request/url",
        }

        mocked_exception_handler.assert_not_called()
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 500)
        mocked_exception_handler.assert_called()

    def test_wsgi_script_binary_support_with_content_encoding(self):
        """
        Ensure that response body is base64 encoded when BINARY_SUPPORT is enabled and Content-Encoding header is present.
        """
        lh = LambdaHandler("tests.test_binary_support_settings")

        text_plain_event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/content_encoding_header_json1",
        }

        # A likely scenario is that the application would be gzip compressing some json response. That's checked first.
        response = lh.handler(text_plain_event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertIn("isBase64Encoded", response)
        self.assertTrue(is_base64(response["body"]))

        # We also verify that some unknown mimetype with a Content-Encoding also encodes to b64. This route serves
        # bytes in the response.

        text_arbitrary_event = {
            **text_plain_event,
            **{"path": "/content_encoding_header_textarbitrary1"},
        }

        response = lh.handler(text_arbitrary_event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertIn("isBase64Encoded", response)
        self.assertTrue(is_base64(response["body"]))

        # This route is similar to the above, but it serves its response as text and not bytes. That the response
        # isn't bytes shouldn't matter because it still has a Content-Encoding header.

        application_json_event = {
            **text_plain_event,
            **{"path": "/content_encoding_header_textarbitrary2"},
        }

        response = lh.handler(application_json_event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertIn("isBase64Encoded", response)
        self.assertTrue(is_base64(response["body"]))

    def test_wsgi_script_binary_support_without_content_encoding_edgecases(
        self,
    ):
        """
        Ensure zappa response bodies are NOT base64 encoded when BINARY_SUPPORT is enabled and the mimetype is "application/json" or starts with "text/".
        """
        lh = LambdaHandler("tests.test_binary_support_settings")

        text_plain_event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/textplain_mimetype_response1",
        }

        for path in [
            "/textplain_mimetype_response1",  # text/plain mimetype should not be turned to base64
            "/textarbitrary_mimetype_response1",  # text/arbitrary mimetype should not be turned to base64
            "/json_mimetype_response1",  # application/json mimetype should not be turned to base64
        ]:
            event = {**text_plain_event, "path": path}
            response = lh.handler(event, None)

            self.assertEqual(response["statusCode"], 200)
            self.assertNotIn("isBase64Encoded", response)
            self.assertFalse(is_base64(response["body"]))

    def test_wsgi_script_binary_support_without_content_encoding(
        self,
    ):
        """
        Ensure zappa response bodies are base64 encoded when BINARY_SUPPORT is enabled and Content-Encoding is absent.
        """
        lh = LambdaHandler("tests.test_binary_support_settings")

        text_plain_event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/textplain_mimetype_response1",
        }

        for path in [
            "/arbitrarybinary_mimetype_response1",
            "/arbitrarybinary_mimetype_response2",
        ]:
            event = {**text_plain_event, "path": path}
            response = lh.handler(event, None)

            self.assertEqual(response["statusCode"], 200)
            self.assertIn("isBase64Encoded", response)
            self.assertTrue(is_base64(response["body"]))

    def test_wsgi_script_binary_support_userdefined_additional_text_mimetypes__defined(
        self,
    ):
        """
        Ensure zappa response bodies are NOT base64 encoded when BINARY_SUPPORT is True, and additional_text_mimetypes are defined
        """
        lh = LambdaHandler("tests.test_binary_support_additional_text_mimetypes_settings")
        expected_additional_mimetypes = ["application/vnd.oai.openapi"]
        self.assertEqual(lh.settings.ADDITIONAL_TEXT_MIMETYPES, expected_additional_mimetypes)

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/userdefined_additional_mimetype_response1",
        }

        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertNotIn("isBase64Encoded", response)
        self.assertFalse(is_base64(response["body"]))

    def test_wsgi_script_binary_support_userdefined_additional_text_mimetypes__undefined(
        self,
    ):
        """
        Ensure zappa response bodies are base64 encoded when BINARY_SUPPORT is True and mimetype not defined in additional_text_mimetypes
        """
        lh = LambdaHandler("tests.test_binary_support_settings")

        event = {
            "body": "",
            "resource": "/{proxy+}",
            "requestContext": {},
            "queryStringParameters": {},
            "headers": {
                "Host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "pathParameters": {"proxy": "return/request/url"},
            "httpMethod": "GET",
            "stageVariables": {},
            "path": "/userdefined_additional_mimetype_response1",
        }

        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertIn("isBase64Encoded", response)
        self.assertTrue(is_base64(response["body"]))

    def test_wsgi_script_on_cognito_event_request(self):
        """
        Ensure that requests sent by cognito behave sensibly
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "version": "1",
            "region": "eu-west-1",
            "userPoolId": "region_poolID",
            "userName": "uuu-id-here",
            "callerContext": {
                "awsSdkVersion": "aws-sdk-js-2.149.0",
                "clientId": "client-id-here",
            },
            "triggerSource": "PreSignUp_SignUp",
            "request": {
                "userAttributes": {"email": "email@example.com"},
                "validationData": None,
            },
            "response": {
                "autoConfirmUser": False,
                "autoVerifyEmail": False,
                "autoVerifyPhone": False,
            },
        }

        response = lh.handler(event, None)

        self.assertEqual(response["response"]["autoConfirmUser"], False)

    def test_bot_triggered_event(self):
        """
        Ensure that bot triggered events are handled as in the settings
        """
        lh = LambdaHandler("tests.test_bot_handler_being_triggered")
        # from : https://docs.aws.amazon.com/lambda/latest/dg/eventsources.html#eventsources-lex
        event = {
            "messageVersion": "1.0",
            "invocationSource": "DialogCodeHook",
            "userId": "user-id specified in the POST request to Amazon Lex.",
            "sessionAttributes": {
                "key1": "value1",
                "key2": "value2",
            },
            "bot": {"name": "bot-name", "alias": "bot-alias", "version": "bot-version"},
            "outputDialogMode": "Text or Voice, based on ContentType request header in runtime API request",
            "currentIntent": {
                "name": "intent-name",
                "slots": {
                    "slot-name": "value",
                    "slot-name": "value",
                    "slot-name": "value",
                },
                "confirmationStatus": "None, Confirmed, or Denied (intent confirmation, if configured)",
            },
        }

        response = lh.handler(event, None)

        self.assertEqual(response, "Success")

    def test_exception_in_bot_triggered_event(self):
        """
        Ensure that bot triggered exceptions are handled as defined in the settings.
        """
        lh = LambdaHandler("tests.test_bot_exception_handler_settings")
        # from : https://docs.aws.amazon.com/lambda/latest/dg/eventsources.html#eventsources-lex
        event = {
            "messageVersion": "1.0",
            "invocationSource": "DialogCodeHook",
            "userId": "user-id specified in the POST request to Amazon Lex.",
            "sessionAttributes": {
                "key1": "value1",
                "key2": "value2",
            },
            "bot": {"name": "bot-name", "alias": "bot-alias", "version": "bot-version"},
            "outputDialogMode": "Text or Voice, based on ContentType request header in runtime API request",
            "currentIntent": {
                "name": "intent-name",
                "slots": {
                    "slot-name": "value",
                    "slot-name": "value",
                    "slot-name": "value",
                },
                "confirmationStatus": "None, Confirmed, or Denied (intent confirmation, if configured)",
            },
        }

        response = lh.lambda_handler(event, None)
        mocked_exception_handler.assert_called

    def test_wsgi_script_name_on_alb_event(self):
        """
        Ensure ALB-triggered events are properly handled by LambdaHandler
        ALB-forwarded events have a slightly different request structure than API-Gateway
        https://docs.aws.amazon.com/elasticloadbalancing/latest/application/lambda-functions.html
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "requestContext": {
                "elb": {
                    "targetGroupArn": "arn:aws:elasticloadbalancing:region:123456789012:targetgroup/my-target-group/6d0ecf831eec9f09"
                }
            },
            "httpMethod": "GET",
            "path": "/return/request/url",
            "queryStringParameters": {},
            "headers": {
                "accept": "text/html,application/xhtml+xml",
                "accept-language": "en-US,en;q=0.8",
                "content-type": "text/plain",
                "cookie": "cookies",
                "host": "1234567890.execute-api.us-east-1.amazonaws.com",
                "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6)",
                "x-amzn-trace-id": "Root=1-5bdb40ca-556d8b0c50dc66f0511bf520",
                "x-forwarded-for": "72.21.198.66",
                "x-forwarded-port": "443",
                "x-forwarded-proto": "https",
            },
            "isBase64Encoded": False,
            "body": "",
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        self.assertEqual(response["statusDescription"], "200 OK")
        self.assertEqual(response["isBase64Encoded"], False)
        self.assertEqual(
            response["body"],
            "https://1234567890.execute-api.us-east-1.amazonaws.com/return/request/url",
        )

    def test_merge_headers_no_multi_value(self):
        event = {"headers": {"a": "b"}}

        merged = merge_headers(event)
        self.assertEqual(merged["a"], "b")

    def test_merge_headers_combine_values(self):
        event = {
            "headers": {"a": "b", "z": "q"},
            "multiValueHeaders": {"a": ["c"], "x": ["y"]},
        }

        merged = merge_headers(event)
        self.assertEqual(merged["a"], "c")
        self.assertEqual(merged["x"], "y")
        self.assertEqual(merged["z"], "q")

    def test_merge_headers_no_single_value(self):
        event = {"multiValueHeaders": {"a": ["c", "d"], "x": ["y", "z", "f"]}}
        merged = merge_headers(event)
        self.assertEqual(merged["a"], "c, d")
        self.assertEqual(merged["x"], "y, z, f")

    def test_cloudwatch_subscription_event(self):
        """
        Test that events sent in the format used by CloudWatch logs via
        subscription filters are handled properly.
        The actual payload that Lambda receives is in the following format
        { "awslogs": {"data": "BASE64ENCODED_GZIP_COMPRESSED_DATA"} }
        https://docs.aws.amazon.com/AmazonCloudWatch/latest/logs/SubscriptionFilters.html
        """
        lh = LambdaHandler("tests.test_event_script_settings")

        event = {"awslogs": {"data": "some-data-not-important-for-test"}}
        response = lh.handler(event, None)

        self.assertEqual(response, True)

    def test_wsgi_script_name_on_v2_formatted_event_function_url(self):
        """
        Ensure that requests with payload format version 2.0 succeed (Function URL case - no stage)
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "version": "2.0",
            "routeKey": "$default",
            "rawPath": "/return/request/url",
            "rawQueryString": "",
            "headers": {
                "host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "requestContext": {
                "http": {
                    "method": "GET",
                    "path": "/return/request/url",
                },
                # No "stage" key - this is a Function URL, not API Gateway v2
            },
            "isBase64Encoded": False,
            "body": "",
            "cookies": ["Cookie_1=Value1; Expires=21 Oct 2021 07:48 GMT", "Cookie_2=Value2; Max-Age=78000"],
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        # Function URL has no stage, so URL should not include /dev
        self.assertEqual(
            response["body"],
            "https://1234567890.execute-api.us-east-1.amazonaws.com/return/request/url",
        )

    def test_wsgi_script_name_on_v2_formatted_event_with_stage_in_path(self):
        """
        Reproduces issue #1389: API Gateway v2 includes stage name in rawPath
        When API Gateway v2 sends rawPath="/dev/collection" for stage "dev",
        the stage prefix should be stripped to get PATH_INFO="/collection"
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "version": "2.0",
            "routeKey": "$default",
            "rawPath": "/dev/return/request/url",  # Stage "dev" is included in rawPath
            "rawQueryString": "",
            "headers": {
                "host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "requestContext": {
                "http": {
                    "method": "GET",
                    "path": "/dev/return/request/url",
                },
                "stage": "dev",
            },
            "isBase64Encoded": False,
            "body": "",
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        # Expected: stage prefix is stripped, so we get /dev/return/request/url
        # (the /dev in SCRIPT_NAME + /return/request/url in PATH_INFO)
        self.assertEqual(
            response["body"],
            "https://1234567890.execute-api.us-east-1.amazonaws.com/dev/return/request/url",
        )

    def test_wsgi_environ_fixes_double_prefix_bug(self):
        """
        Test that verifies the fix for issue #1389 (double prefix bug).
        This test calls the /debug/wsgi/environ endpoint to inspect SCRIPT_NAME and PATH_INFO.

        Issue #1389: With API Gateway v2, rawPath="/dev/debug/wsgi/environ" includes the stage.
        The fix correctly strips the stage prefix from PATH_INFO.
        SCRIPT_NAME="/dev" and PATH_INFO="/debug/wsgi/environ" (stage stripped)
        Flask correctly builds: /dev + /debug/wsgi/environ = /dev/debug/wsgi/environ
        """
        lh = LambdaHandler("tests.test_wsgi_script_name_settings")

        event = {
            "version": "2.0",
            "routeKey": "$default",
            "rawPath": "/dev/debug/wsgi/environ",  # Stage "dev" is included in rawPath
            "rawQueryString": "",
            "headers": {
                "host": "1234567890.execute-api.us-east-1.amazonaws.com",
            },
            "requestContext": {
                "http": {
                    "method": "GET",
                    "path": "/dev/debug/wsgi/environ",
                },
                "stage": "dev",
            },
            "isBase64Encoded": False,
            "body": "",
        }
        response = lh.handler(event, None)

        self.assertEqual(response["statusCode"], 200)
        # The response body shows the WSGI environ values
        # FIXED behavior: SCRIPT_NAME='/dev' PATH_INFO='/debug/wsgi/environ' (stage stripped correctly)
        # Flask correctly builds URL: /dev + /debug/wsgi/environ = /dev/debug/wsgi/environ
        print(f"\nDEBUG - Response body: {response['body']}")

        # Verify the fix: stage is stripped from PATH_INFO
        self.assertIn("SCRIPT_NAME='/dev'", response["body"])
        self.assertIn("PATH_INFO='/debug/wsgi/environ'", response["body"])  # FIXED: stage stripped
        self.assertIn("/dev/debug/wsgi/environ", response["body"])  # Correct single prefix!



API_STAGE = "dev"
APP_FUNCTION = "app"
APP_MODULE = "tests.test_wsgi_binary_support_app"
BINARY_SUPPORT = True
CONTEXT_HEADER_MAPPINGS = {}
DEBUG = "True"
DJANGO_SETTINGS = None
DOMAIN = "api.example.com"
ENVIRONMENT_VARIABLES = {}
LOG_LEVEL = "DEBUG"
PROJECT_NAME = "binary_support_settings"
COGNITO_TRIGGER_MAPPING = {}
EXCEPTION_HANDLER = None
ADDITIONAL_TEXT_MIMETYPES = ["application/vnd.oai.openapi"]



API_STAGE = "dev"
APP_FUNCTION = "raises_exception"
APP_MODULE = "tests.test_handler"
BINARY_SUPPORT = False
CONTEXT_HEADER_MAPPINGS = {}
DEBUG = "True"
DJANGO_SETTINGS = None
DOMAIN = "api.example.com"
ENVIRONMENT_VARIABLES = {}
LOG_LEVEL = "DEBUG"
PROJECT_NAME = "raises_exception"
COGNITO_TRIGGER_MAPPING = {}
EXCEPTION_HANDLER = "tests.test_handler.mocked_exception_handler"



# -*- coding: utf8 -*-
import sys
import unittest
from io import BytesIO

from zappa.middleware import ZappaWSGIMiddleware, all_casings
from zappa.wsgi import create_wsgi_request


class TestWSGIMockMiddleWare(unittest.TestCase):
    """
    These tests can cheat and have access to the inner status and headers,
    through _start_response.
    """

    def setUp(self):
        """
        Set the test up with default headers and status codes.
        """
        self.headers = list()
        self.status = list()

    def _start_response(self, status, headers, exc_info=None):
        self.status[:] = [status]
        self.headers[:] = headers

    def test_all_casings(self):
        # 2^9
        input_string = "Set-Cookie"
        x = 0
        for casing in all_casings(input_string):
            x = x + 1
        self.assertEqual(x, 512)

        # 2^0
        input_string = ""
        x = 0
        for casing in all_casings(input_string):
            x = x + 1
        self.assertEqual(x, 1)

    def test_wsgi_middleware_uglystring(self):
        ugly_string = "˝ÓÔÒÚÆ☃ЗИЙКЛМФХЦЧШ차를 타고 온 펲시맨(╯°□°）╯︵ ┻━┻)"

        # Pass some unicode through the middleware body
        def simple_app(environ, start_response):
            # String of weird characters
            status = "200 OK"
            response_headers = []
            start_response(status, response_headers)
            return [ugly_string]

        # Wrap the app with the middleware
        app = ZappaWSGIMiddleware(simple_app)

        # Call with empty WSGI Environment
        resp = app(dict(), self._start_response)
        print("".join(resp))

        # Pass some unicode through the middleware headers
        def simple_app(environ, start_response):
            # String of weird characters
            status = "301 Moved Permanently"
            response_headers = [("Location", f"http://zappa.com/elsewhere{ugly_string}")]
            start_response(status, response_headers)
            return [ugly_string]

        # Wrap the app with the middleware
        app = ZappaWSGIMiddleware(simple_app)

        # Call with empty WSGI Environment
        resp = app(dict(), self._start_response)
        print("".join(resp))

    def test_wsgi_authorizer_handling(self):
        # With user
        event = {
            "httpMethod": "GET",
            "queryStringParameters": None,
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {"authorizer": {"principalId": "user1"}},
            "query": {},
        }

        environ = create_wsgi_request(event, script_name="http://zappa.com/", trailing_slash=False)
        self.assertEqual(environ["REMOTE_USER"], "user1")

        # With empty authorizer, should not include REMOTE_USER
        event = {
            "httpMethod": "GET",
            "queryStringParameters": None,
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {"authorizer": {"principalId": ""}},
            "query": {},
        }

        environ = create_wsgi_request(event, script_name="http://zappa.com/", trailing_slash=False)
        user = environ.get("REMOTE_USER", "no_user")
        self.assertEqual(user, "no_user")

        # With missing authorizer, should not include REMOTE_USER
        event = {
            "httpMethod": "GET",
            "queryStringParameters": None,
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {},
            "query": {},
        }

        environ = create_wsgi_request(event, script_name="http://zappa.com/", trailing_slash=False)
        user = environ.get("REMOTE_USER", "no_user")
        self.assertEqual(user, "no_user")

        # With empty authorizer, should not include REMOTE_USER
        event = {
            "httpMethod": "GET",
            "queryStringParameters": None,
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {"authorizer": {}},
            "query": {},
        }

        environ = create_wsgi_request(event, script_name="http://zappa.com/", trailing_slash=False)
        user = environ.get("REMOTE_USER", "no_user")
        self.assertEqual(user, "no_user")

    def test_wsgi_map_context_headers_handling(self):
        # Validate a single context value mapping is translated into a HTTP header
        event = {
            "httpMethod": "GET",
            "queryStringParameters": None,
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {
                "authorizer": {"principalId": "user1"},
            },
            "query": {},
        }

        environ = create_wsgi_request(
            event,
            script_name="http://zappa.com/",
            trailing_slash=False,
            context_header_mappings={"PrincipalId": "authorizer.principalId"},
        )
        self.assertEqual(environ["HTTP_PRINCIPALID"], "user1")

        # Validate multiple mappings with an invalid mapping
        # Invalid mapping should be ignored
        event = {
            "httpMethod": "GET",
            "queryStringParameters": None,
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {
                "resourceId": "123456",
                "apiId": "1234567890",
                "resourcePath": "/{proxy+}",
                "httpMethod": "POST",
                "requestId": "c6af9ac6-7b61-11e6-9a41-93e8deadbeef",
                "accountId": "123456789012",
                "identity": {
                    "userAgent": "Custom User Agent String",
                    "cognitoIdentityPoolId": "userpoolID",
                    "cognitoIdentityId": "myCognitoID",
                    "sourceIp": "127.0.0.1",
                },
                "stage": "prod",
            },
            "query": {},
        }

        environ = create_wsgi_request(
            event,
            script_name="http://zappa.com/",
            trailing_slash=False,
            context_header_mappings={
                "CognitoIdentityID": "identity.cognitoIdentityId",
                "APIStage": "stage",
                "InvalidValue": "identity.cognitoAuthenticationType",
                "OtherInvalid": "nothinghere",
            },
        )
        self.assertEqual(environ["HTTP_COGNITOIDENTITYID"], "myCognitoID")
        self.assertEqual(environ["HTTP_APISTAGE"], "prod")
        self.assertNotIn("HTTP_INVALIDVALUE", environ)
        self.assertNotIn("HTTP_OTHERINVALID", environ)

    def test_wsgi_input_as_file_like_object(self):
        event = {
            "httpMethod": "GET",
            "path": "/v1/runs",
            "body": None,
        }

        environ = create_wsgi_request(event, script_name="http://zappa.com/", trailing_slash=False)
        self.assertEqual(type(environ["wsgi.input"]), BytesIO)

    def test_should_allow_empty_query_params(self):
        event = {
            "httpMethod": "GET",
            "queryStringParameters": {},
            "multiValueQueryStringParameters": {},
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {
                "resourceId": "123456",
                "apiId": "1234567890",
                "resourcePath": "/{proxy+}",
                "httpMethod": "POST",
                "requestId": "c6af9ac6-7b61-11e6-9a41-93e8deadbeef",
                "accountId": "123456789012",
                "identity": {
                    "userAgent": "Custom User Agent String",
                    "cognitoIdentityPoolId": "userpoolID",
                    "cognitoIdentityId": "myCognitoID",
                    "sourceIp": "127.0.0.1",
                },
                "stage": "prod",
            },
            "query": {},
        }
        environ = create_wsgi_request(event, script_name="http://zappa.com/", trailing_slash=False)
        self.assertEqual(environ["QUERY_STRING"], "")

    def test_should_handle_multi_value_query_string_params(self):
        event = {
            "httpMethod": "GET",
            "queryStringParameters": {},
            "multiValueQueryStringParameters": {"foo": [1, 2]},
            "path": "/v1/runs",
            "params": {},
            "body": None,
            "headers": {"Content-Type": "application/json"},
            "pathParameters": {"proxy": "v1/runs"},
            "requestContext": {
                "resourceId": "123456",
                "apiId": "1234567890",
                "resourcePath": "/{proxy+}",
                "httpMethod": "POST",
                "requestId": "c6af9ac6-7b61-11e6-9a41-93e8deadbeef",
                "accountId": "123456789012",
                "identity": {
                    "userAgent": "Custom User Agent String",
                    "cognitoIdentityPoolId": "userpoolID",
                    "cognitoIdentityId": "myCognitoID",
                    "sourceIp": "127.0.0.1",
                },
                "stage": "prod",
            },
            "query": {},
        }
        environ = create_wsgi_request(event, script_name="http://zappa.com/", trailing_slash=False)
        self.assertEqual(environ["QUERY_STRING"], "foo=1&foo=2")



import os
import sys

# add the Lambda root path into the sys.path
sys.path.append("/var/task")


def get_django_wsgi(settings_module):
    from django.core.wsgi import get_wsgi_application

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings_module)

    return get_wsgi_application()



from flask import Flask, request

app = Flask(__name__)


@app.route("/return/request/url", methods=["GET", "POST"])
def return_request_url():
    return request.url


@app.route("/debug/wsgi/environ", methods=["GET"])  # Route without stage prefix (Flask sees clean path)
def debug_wsgi_environ():
    """Debug endpoint to inspect WSGI environ values"""
    script_name = request.environ.get("SCRIPT_NAME", "")
    path_info = request.environ.get("PATH_INFO", "")
    return f"SCRIPT_NAME={script_name!r} PATH_INFO={path_info!r} URL={request.url}"



import os
import sys


def running_in_docker() -> bool:
    """
    Determine if zappa is running in docker.
    - When docker is used allow usage of any python version
    """
    # https://stackoverflow.com/questions/63116419
    running_in_docker_flag = os.getenv("ZAPPA_RUNNING_IN_DOCKER", "False").lower() in {"y", "yes", "t", "true", "1"}
    return running_in_docker_flag


SUPPORTED_VERSIONS = [(3, 9), (3, 10), (3, 11), (3, 12), (3, 13)]
MINIMUM_SUPPORTED_MINOR_VERSION = 9


if not running_in_docker() and sys.version_info[:2] not in SUPPORTED_VERSIONS:
    print(running_in_docker())
    formatted_supported_versions = ["{}.{}".format(*version) for version in SUPPORTED_VERSIONS]
    err_msg = "This version of Python ({}.{}) is not supported!\n".format(
        *sys.version_info
    ) + "Zappa (and AWS Lambda) support the following versions of Python: {}".format(formatted_supported_versions)
    raise RuntimeError(err_msg)
elif running_in_docker() and sys.version_info.minor < MINIMUM_SUPPORTED_MINOR_VERSION:
    # when running in docker enforce minimum version only
    err_msg = (
        f"This version of Python ({sys.version_info.major}.{sys.version_info.minor}) is not supported!\n"
        f"Zappa requires a minimum version of 3.{MINIMUM_SUPPORTED_MINOR_VERSION}"
    )
    raise RuntimeError(err_msg)

__version__ = "0.61.1"



API_STAGE = "dev"
APP_FUNCTION = "handler_for_events"
APP_MODULE = "tests.test_event_script_app"
DEBUG = "True"
DJANGO_SETTINGS = None
DOMAIN = "api.example.com"
ENVIRONMENT_VARIABLES = {}
LOG_LEVEL = "DEBUG"
PROJECT_NAME = "test_event_script_app"
COGNITO_TRIGGER_MAPPING = {}



API_STAGE = "dev"
APP_FUNCTION = "app"
APP_MODULE = "tests.test_wsgi_script_name_app"
BINARY_SUPPORT = False
CONTEXT_HEADER_MAPPINGS = {}
DEBUG = "True"
DJANGO_SETTINGS = None
DOMAIN = "api.example.com"
ENVIRONMENT_VARIABLES = {}
LOG_LEVEL = "DEBUG"
PROJECT_NAME = "wsgi_script_name_settings"
COGNITO_TRIGGER_MAPPING = {}
AWS_BOT_EVENT_MAPPING = {"intent-name:DialogCodeHook": "tests.test_handler.handle_bot_intent"}
